;; Generated package description from tmr.el  -*- no-byte-compile: t -*-
(define-package "tmr" "0.4.0" "Set timers using a convenient notation" '((emacs "27.1")) :commit "0d6ce6979c980f0d61c61d1f6478f805612315b4" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com") ("Damien Cassou" . "damien@cassou.me") ("Daniel Mendler" . "mail@daniel-mendler.de")) :maintainer '("TMR Development" . "~protesilaos/tmr@lists.sr.ht") :keywords '("convenience" "timer") :url "https://git.sr.ht/~protesilaos/tmr")
